﻿/*
 * Token.cs - Token Regex
 * 
 * ShakeIt - Mutation engine for fuzzing
 * Jeremy Brown (jbrown3264/gmail)
 * Nov 2015
 * 
 */

using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;

namespace ShakeIt
{
    public class Token
    {
        /*
         * Determine if we've matched a token from input and if so, add it's match.position (Index)
         * and length (match.Value.Length) to our list of objects
         */
        public static void Match(string input, List<TokenData> tokens, string regex)
        {
            foreach (Match match in Regex.Matches(input, regex))
            {
                string newToken = input.Substring(match.Index, match.Value.Length);

                if (!Program.BlackList.Any(newToken.Equals))
                {
                    tokens.Add(new TokenData(match.Index, match.Value.Length));
                }
            }
        }


        /*
         * Use regex to match various tokens we're interested in shaking
         */
        public static List<TokenData> Find(char token, string input)
        {
            List<TokenData> tokens = new List<TokenData>();
            bool skip = false;

            switch (token)
            {
                /*
                 * JS/DOM Function
                 * matches 'myFunction' in function myFunction()
                 * 
                 * Reference: http://stackoverflow.com/questions/11500194/match-functions-and-function-calls-in-c-using-regex
                 */
                case '(':
                    //
                    // [negative lookahead (no loops)]    [negative look behind for '.' (no methods)] [positive lookahead for '(' (only functions)]
                    // (?!\b(if|while|for)\b)             (?<!\.) \b\w+                               (?=\()
                    //
                    Match(input, tokens, @"(?!\b(if|while|for)\b)(?<!\.)\b\w+(?=\()");
                    break;

                /*
                 * CSS Property
                 * matches 'font-size' in CSS font-size: 10px; 
                 */
                case ':':
                    Match(input, tokens, @"(?!\\b)\b[\w-]+(?=\S*\:)");
                    break;

                /*
                 * HTML Attribute
                 * matches 'class' in <h1 class="center">
                 */
                case '=':
                    Match(input, tokens, @"(?!\\b)\b[\w-]+(?=\S*\=)");
                    break;

                 /*
                 * HTML/JS/DOM Value
                 * matches 'demo1' in <p id="demo1"></p>
                 */
                case '"':
                    Match(input, tokens, "(?<==\")(?:(\\w|\\(.*\\))+)");
                    break;

                /*
                 * HTML Tag
                 * matches 'button' in <button onclick="myFunction()">Try it</button>
                 * 
                 */
                case '<':
                    Match(input, tokens, "(?<=<|\\/)(?:(\\w|\\(.*\\))+)");
                    break;

                /*
                 * JS/DOM Method
                 * matches 'getElementById' in document.getElementById("demo1");
                 */
                case '.':
                    Match(input, tokens, @"(?<=\.)\b\w+(?=\()");
                    //Match(input, tokens, @"(?<=\.)\b\w+\\(?=\(\\)");
                    break;

                /*
                 * JS/DOM Variable
                 * matches 'str' in var str = "test";
                 */
                case 'v':
                    Match(input, tokens, @"(?<=var\s)(?:(\w+))");
                    break;

                /*
                 * JS/DOM Property
                 * matches 'target' in document.getElementById("demo2").target = "_blank";
                 */
                case 'p':
                    //
                    // [positive lookbehind for '.'] [find any words] [positive lookahead for '=']
                    // (?<=\.)                       \w+             (?=\s\=)
                    //

                    Match(input, tokens, @"(?<=\.)\w+(?=\s\=)");
                    break;

                /*
                 * JS/DOM Parameters
                 * matches "demo1" in document.getElementById("demo1").innerHTML = n;
                 * and
                 * matches 'event.changeEx' in parseInt\(event.changeEx\) (eg. PDF)
                 */
                case ')':
                    //Match(input, tokens, "(?<=\\()(?:(\"\\w+\\\"|\\w+))");
                    Match(input, tokens, "(?<=\\()(?:(\"*\\w+\"*[^\\)]+|\\w+))");
                    break;
            }

            return tokens;
        }
    }
}
